
#define _BSD_SOURCE // requered by gethostname() to work correctly

//~ #define DEBUG

#include <unistd.h>
#include <stdlib.h>
#include"mnb-utils.h"
#include"mnb-environment.h"

#ifndef HOST_NAME_MAX
#define HOST_NAME_MAX 64 // linux standard according to man page for gethostname
#endif



/**
 * provides a list of the hostnames of all ranks
 **/
static char * MNB_GetAllHostNames();

/**
 * adds the rank to the node with the provided hostname
 * if no such node can be found a new one is created 
 **/
static void MNB_AddRankToEnvironment(MNB_Environment * env, int rank, char * hostname);

/**
 * returns the node, on which the current Rank is running
 **/
static void MNB_FindLocalNode(MNB_Environment * env);


/**
 * select a node and a rank and sets them as schedulerRank and schedulerNode
 **/
static void MNB_PickScheduler(MNB_Environment * env);

/**
 * allocates the requered memory and return the hostname by calling
 * gethostname()
 * 
 **/
static char * MNB_GetHostname();


/**
 * Allocates the memory and sets default values for an MNB_Environment
 * 
 **/
MNB_Environment * MNB_CreateEnvironment(){
	MNB_Environment * newEnvironment = malloc(sizeof(MNB_Environment));
	newEnvironment->numberOfNodes = 0;
	newEnvironment->numberOfRanks = 0;
	newEnvironment->schedulerRank = 0;
	
	newEnvironment->schedulerNode	= NULL;
	newEnvironment->localNode		= NULL;
	
	newEnvironment->testerNodes 	= MNB_CreateList();
	newEnvironment->singleNodeExecution = FALSE;
	return newEnvironment;
}

/**
 *create an shallow copy of an MNB_Environment
 * 
 **/
MNB_Environment * MNB_CopyEnvironment(MNB_Environment * origEnvironment)
{
	//simple copies of all values
	MNB_Environment * newEnvironment = MNB_CreateEnvironment();
	newEnvironment->numberOfNodes = origEnvironment->numberOfNodes;
	newEnvironment->numberOfRanks = origEnvironment->numberOfRanks;
	newEnvironment->schedulerRank = origEnvironment->schedulerRank;
	
	newEnvironment->schedulerNode	= origEnvironment->schedulerNode;
	newEnvironment->localNode		= origEnvironment->localNode;
	
	newEnvironment->testerNodes 	= MNB_CopyList(origEnvironment->testerNodes);
	
	newEnvironment->singleNodeExecution = origEnvironment->singleNodeExecution;
	return newEnvironment;
}


/**
 * 
 * allocates the momory and sets default values for an node
 * 
 **/
MNB_Node * MNB_CreateNode()
{
	MNB_Node * newNode	= malloc(sizeof(MNB_Node));
	newNode->hostname	= NULL;
	newNode->ranks		= MNB_CreateList();
	newNode->status		= MNB_NODE_FREE;
	return newNode;
}


/**
 * frees up the memory used to contain the node
 * 
 **/
void MNB_DeleteNode(MNB_Node * node)
{
	free(node->hostname);
	
	//deleting all ranks working on this node
	//delete list would only free the list references to them, but not
	//the memory requered by the ranks 
	struct MNB_ListElement * iterator = node->ranks->first;
	while(iterator != NULL){
		MNB_RemoveFromList(iterator->data, node->ranks);
		free(iterator->data);
		iterator = iterator->next;
		}
	//and now can delete the list
	MNB_DeleteList(node->ranks);
	
	
	free(node);
}


/**
 *  frees up the memory used to contain the environment
 * 
 **/
void MNB_DeleteEnvironment(MNB_Environment * env)
{
	MNB_println("MNB_DeleteEnvironment");
	//since localNode is just a reference to a node contained in the testerNode list
	//it will be deleted, when we go throu the list. The same is valid for the
	//schduler Node, but only if we are using singleNodeExecution
	if(!env->singleNodeExecution){
		MNB_DeleteNode(env->schedulerNode);
	} 
	
	//and now are freeing up every node we have left
	struct MNB_ListElement * iterator = env->testerNodes->first;
	while(iterator != NULL)
	{
		MNB_RemoveFromList(iterator->data, env->testerNodes);
		MNB_DeleteNode(iterator->data);
		iterator = iterator->next;
	}
	
	
	
	
	free(env);
	MNB_println("MNB_DeleteEnvironment:done");
}


/**
 * goes to the list of slaveHosts to find the one, with the least
 * resources (ranks) available, so es few as possible gets waisted.
 * 
 * once its through with the list it assigns the best candidate to be
 * schedulerNode and its first rank to be schedulerRank
 * 
 **/
void MNB_PickScheduler(MNB_Environment * env)
{
	MNB_println("MNB_PickScheduler");
	assert(env != NULL);
	assert(env->testerNodes != NULL);
	
	
	//the node with the lowest number of ranks running on it is becoming schedulerNode
	
	
	//we can start we any, so why not take the first one
	env->schedulerNode = (MNB_Node *)env->testerNodes->first->data;
	
	//now we just need to go through the list and ceep a pointer to the smalles node
	struct MNB_ListElement * iterator = env->testerNodes->first;
	while(iterator != NULL){
		MNB_Node * host = (MNB_Node* )iterator->data;
		if(host->ranks->size < env->schedulerNode->ranks->size){
				env->schedulerNode = host;
		}
		iterator = iterator->next;
	}
	
	//again, we could user any of the ranks on scheduler Node, so we take the first one
	//and we need just the value, not the memory allocated for it before
	// we have to remove the schedulerRank from the schedulerNode list
	// so we can use the node in single Node execution
	// we have no other reference but the list entry we are about to delete
	int * tmpSchedulerRank = env->schedulerNode->ranks->first->data; 
	
	MNB_RemoveFromList(tmpSchedulerRank, env->schedulerNode->ranks);
	
	env->schedulerRank = *tmpSchedulerRank;
	//now we can get rid of the memory
	free(tmpSchedulerRank);


	//workaround to make this usable on one node;
	if(env->testerNodes->size > 1){ //usualy we would't want the schedulerNode to appear in the list at all
	//but if we work with only one host, we can't just through it out of the list
	//like we do now
	MNB_RemoveFromList(env->schedulerNode, env->testerNodes);
	
	} 
	
	MNB_println("MNB_PickScheduler:done");
}

/**
 * fills the given enviroment with information about the actual MPI enviroment;
 * expects MPI to be allready initialised.
 * 
 **/
MNB_Environment * MNB_ReadEnvironment(MNB_Environment * env){
	int worldSize;

	MPI_Comm_size(MPI_COMM_WORLD, &worldSize);
	
	//we need the hostnames of the all ranks so we can find out, which
	//ones are working on the same node
	char * allHostNames = MNB_GetAllHostNames();
        		
    //now call MNB_AddRankToEnvironment for each rank and its hostname,
    //so they can be placed right within the environment
    for (int i = 0; i < worldSize; i++){
		//goint through the host name list
		char * current =  &allHostNames[(HOST_NAME_MAX + 1) * i];
		MNB_AddRankToEnvironment(env, i, current);
	}
	
	//now we know the environment there is no user for the hostnames      
	free(allHostNames);


	//if we have just one Node -> its singleNodeExecution
	if(env->testerNodes->size == 1){
		env->singleNodeExecution = TRUE;
		MNB_printf("%s\n", "singleHostExecution == True");
	} else {
		env->singleNodeExecution = FALSE;
		MNB_printf("%s\n", "singleHostExecution == False");
	}	

   
   //setting the pointers to special Nodes
   MNB_FindLocalNode(env);
   MNB_PickScheduler(env);
   
   return env;
		
}


/**
 * gets own hostname and exchanges the hostnames with all other ranks
 * by using MPI_Allgather
 **/
char * MNB_GetAllHostNames()
{
	MNB_println("MNB_GetAllHostNames");
	//everyone has to allocate the memory
	char * hostname = MNB_GetHostname();
	
	int numberOfRanks;
	MPI_Comm_size(MPI_COMM_WORLD, &numberOfRanks);
	
	//getting enough memory to store all the hostnames
	char * allHostNames = malloc((HOST_NAME_MAX + 1) * numberOfRanks);
	if (allHostNames == NULL){
		MNB_Abort("Error while reserving memory for recv buffer: %d\n");
        }
    memset(allHostNames, 0, numberOfRanks * (HOST_NAME_MAX+1));
	MNB_println("about to exchange the hostnames");
	//exchange of the hostnames
	MPI_Allgather(hostname, (HOST_NAME_MAX+1), MPI_CHAR, allHostNames,
        		(HOST_NAME_MAX+1), MPI_CHAR, MPI_COMM_WORLD);
        		
       
    //we don't ned our original hostnamemenory anymore  
    free(hostname);
    
       
	return allHostNames;
	
}
/**
 * checks if there is allready a node with the given hostname in the environment.
 * if one can be found, the rank is added to this node, 
 * otherwise a new node is created and added for the given rank
 * 
 **/
void MNB_AddRankToEnvironment(MNB_Environment * env, int rank, char * hostname){
	MNB_printf("MNB_AddRankToEnvironment called with rank:%d, hostname:%s\n", rank, hostname);
	env->numberOfRanks = env->numberOfRanks + 1;
	
	//creating a copy of the rank so we can insert it into lists
	int * newRank = malloc(sizeof(int));
	*newRank = rank;
	
	
	
	int hostFound = FALSE;
	struct MNB_ListElement * iterator = env->testerNodes->first;
	//looking for the name in the list of allready known.
	while(iterator != NULL){
		MNB_Node * node = ((MNB_Node *)(iterator->data));
		if(strcmp(hostname, node->hostname) == 0){
			hostFound = TRUE; // we have found a node, with the same hostname
			MNB_AddToList(newRank, node->ranks); //can not append the rank to it
			break;
		} 
		iterator = iterator->next;
	}
	
	
	//in case none of the nodes had the hostname we were looking for
	//we have to create a new one.
	if(!hostFound){
		env->numberOfNodes = env->numberOfNodes + 1;
		MNB_Node * newNode = MNB_CreateNode();
		
		//taking care of the hostname ( we have to create a copy)
		int hostnameSize = strlen(hostname);
		newNode->hostname = malloc(hostnameSize + 1);
		strcpy(newNode->hostname, hostname);
		newNode->hostname[hostnameSize] = '\0';
		
		//adding the rank to the host
		MNB_AddToList(newRank, newNode->ranks);
		
		
		//and adding the host to the environment
		MNB_AddToList(newNode, env->testerNodes);
	}
}


/**
 * gets the local hostname through gethostname()
 * 
 * lookes through the list of nodes if it can find one with the same name
 * if it can, localNode in env is set
 * 
 **/
void MNB_FindLocalNode(MNB_Environment * env){
	MNB_println("MNB_FindLocalHost");
	
	//we get our own hostname
	char * localHostname = MNB_GetHostname();
	
	//and now we just have to ho through our nodeList and compare the names
	struct MNB_ListElement * iterator = env->testerNodes->first;
	while(iterator != NULL)
	{
		MNB_Node * node = (MNB_Node *)iterator->data;
		MNB_printf("localNode search by name comparison:%s to %s\n",
					localHostname, 
					node->hostname);
		
		if(strcmp(localHostname, node->hostname) == 0){
				env->localNode = node;
				break; 
		}
		iterator = iterator->next;
	}
	//we have no use for the localHostname anymore
	free(localHostname);
}


/**
 * allocates the requered memory and return the hostname by calling
 * gethostname()
 * 
 **/
char * MNB_GetHostname(){
	char * hostname = malloc(HOST_NAME_MAX + 1); //alternative MPI_MAX_PROCESSORNAME
	if (hostname == 0){
		MNB_Abort("Error while reserving memory for my hostname\n");
        }
        
	if(gethostname(hostname, HOST_NAME_MAX) != 0){
		MNB_Abort("couldn't get my own hostname\n");
	}
	return hostname;
}



