


//~ #define DEBUG

#include"mnb-utils.h"
#include"mnb-execution.h"
#include"mnb-environment.h"
#include"mnb-configuration.h"
#include"mnb-scheduling.h"

#include"mnb-stdtests.h"

#include<float.h>



/**
 * for comments on the functions please look at their definitions
 * 
 * 
 **/
static void MNB_ExecuteAsScheduler(MNB_Configuration * conf, MNB_Environment * env);
static void MNB_ExecuteAsTester(MNB_Environment * env);
static void MNB_ExecuteTestFromMessage(MNB_MessageBlock * messageBlock, MPI_Comm comm);
static MNB_MessageBlockBody * MNB_AllocateMessageBlockBody(MNB_MessageBlockHeader * messageHeader);
static void MNB_SendQuitMessageToAll(MNB_Environment * env);
static void MNB_RecvMessage(MNB_Message * message, MNB_Environment * env);
static void MNB_InformTesters(struct MNB_List * nextTestRound, struct MNB_List * nextRoundNodeLists, MNB_Environment * env, MPI_Comm * comm);
static void MNB_ResendMessageToNode(MNB_Message * message, MNB_Node * node);
static void MNB_SendMessage(MNB_Message * message, int targetRank);
static void MNB_FreeMessage(MNB_Message * message);
static void MNB_FillMessageBlock(MNB_MessageBlock * messageBlock, MNB_Test *test, struct MNB_List * nodeList );
static void MNB_SendToResenders( MNB_Message * message, MNB_Environment * env);
static MPI_Comm MNB_CreateCommsFromMessage(MNB_Message * message, MNB_MessageBlock * messageBlock, MPI_Comm * parentComm);



/**
 * The execution separates in this Funktion based on the rolles of the rank
 * (scheduler and tester)
 * the scheduler rank is starting to do the scheduling, while the others are
 * beginning to wait for commands.
 * 
 **/
void MNB_Execute(MNB_Configuration * conf, MNB_Environment * env){
	MNB_println("MNB_Execute:beginnig execution");
	//getting the rank
	int rank;
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
	
	//and checking if we are the scheduler?
 	if(rank == env->schedulerRank){
		MNB_ExecuteAsScheduler(conf, env);
	} else {
		//while the ranks on scheduerNode aren't supposed to execute tests
		//they behave in the same way as the testers, they just don't get any tests
		// to execute 
		MNB_ExecuteAsTester(env);
	}
	MPI_Barrier(MPI_COMM_WORLD);
}




/**
 * tries to get as much as possible of the conf->tests executed, by 
 * sending the appropriate commands to the testerNodes.
 * 
 * because we have to create a custom MPI_Comm for every test, and the creation
 * funktion has to be called by all members of the parent Communicator
 * we have to organise the test execution in rounds, so the testers can
 * create all the comms the will need at the start of every round
 * 
 **/
void MNB_ExecuteAsScheduler(MNB_Configuration * conf, MNB_Environment * env){
	MNB_println("MNB_ExecuteAsScheduler");
	MNB_printf("MNB_ExecuteAsScheduler:beginning scheduling with %d tests\n", conf->tests->size);
	
	
	//sorting the tests and nodes, its nessacery for the scheduling
	conf->tests			= MNB_SortTestList(conf->tests);
	env->testerNodes	= MNB_SortNodeList(env->testerNodes);
	
	
	//all members of MPI_COMM_WORLD have to call this funktion
	//so the testers can create their comm
	MPI_Comm * comm = MNB_CreateTesterComm(env);
	free(comm);
	
	
	//tmp variables we are using in the scheduling loop
	MNB_Test * nextTest = NULL;
	struct MNB_List * nextTestNodeList 		= MNB_CreateList();
	
	//contains the the tests scheduled for the next round
	struct MNB_List * nextTestRound 		= MNB_CreateList();
	//contains the node lists for the tests in the next round in the same order
	struct MNB_List * nextRoundNodeLists	= MNB_CreateList();
	
	int done = FALSE;
	
	while(!done){
		MNB_println("MNB_ExecuteAsScheduler:begin of the scheduler loop");
		
		
		nextTest = MNB_GetFirstSchedulableTest(conf->tests, env, nextTestNodeList);
		//collecting all tests Executable in the next Round
		while(nextTest != NULL){
			#ifdef DEBUG
				MNB_PrintTest(nextTest);
			#endif
			MNB_AddToList(nextTest, nextTestRound);
			MNB_AddToList(nextTestNodeList, nextRoundNodeLists);
			nextTestNodeList = MNB_CreateList();
			//we don't want to schedule this test again
			MNB_RemoveFromList(nextTest, conf->tests);
			MNB_println("MNB_ExecuteAsScheduler:about to start search for an additional test");
			nextTest = MNB_GetFirstSchedulableTest(conf->tests, env, nextTestNodeList);
		}
		//if there are no tests we could execute, we are done
		if(nextTestRound->size == 0){
			done = TRUE;
		//otherwise we have to execute them
		} else {
			MNB_printf("MNB_ExecuteAsScheduler:beginning execution of %d test(s), %d left\n",
						nextTestRound->size, 
						conf->tests->size);
						
			MNB_InformTesters(nextTestRound, nextRoundNodeLists, env, comm);
			//now the work for this round is done, so we clean up to have
			//empty lists for the next one
			MNB_DeleteList(nextTestRound);
			MNB_DeleteList(nextRoundNodeLists);
			nextRoundNodeLists = MNB_CreateList();
			nextTestRound = MNB_CreateList();
			MNB_SetNodesStatus(env->testerNodes, MNB_NODE_FREE);
		}
	
	}
	//we are done with all tests, lets inform everyone about it.
	MNB_SendQuitMessageToAll(env);
	
	//now that we are done, lets clean up
	MNB_DeleteList(nextTestRound);
	MNB_DeleteList(nextTestNodeList);
	MNB_DeleteList(nextRoundNodeLists);

	//sunc up so the following printout will follow all the tests
	MPI_Barrier(MPI_COMM_WORLD);
	
	//in case we could't schedule everything, we schould inform the user
	if(conf->tests->size > 0){
		MNB_printf("%s\n", "the following tests could not be scheduled");
		
		struct MNB_ListElement * notSchedulableTestsIterator = conf->tests->first;
		while(notSchedulableTestsIterator != NULL){
			MNB_Test * notSchedulabeTest = (MNB_Test *)notSchedulableTestsIterator->data;
			MNB_PrintTest(notSchedulabeTest);
			notSchedulableTestsIterator = notSchedulableTestsIterator->next;
			
		}
	}
	
	

}



/**
 * Is executed by the tester ranks to wait for the command from the 
 * schedulerNode and execute them
 * 
 * only returns if the rank recieves a quit message
 * 
 * 
 * blocks the execution by an MPI_Recv call to get the message, interpret
 * it and act accordingly
 * 
 * 
 **/
void MNB_ExecuteAsTester(MNB_Environment * env){
	MNB_println("MNB_ExecuteAsTester");
	MNB_Message message;
	
	//the tester comm is MPI_COMM_WORLD without the scheduler
	//(and the other ranks on schedulerNode if we have more then one node)
	MPI_Comm * testerNodesComm = MNB_CreateTesterComm(env);
	MNB_println("MNB_CreateTesterComm:done");
	
	for(;;){
		MNB_println("is waiting for message");
		MNB_RecvMessage(&message, env);
		MNB_println("recieved a message");
		
		int rank;
		MPI_Comm_rank(MPI_COMM_WORLD, &rank);
		
		
		//to keep the ammount of data a small ass possible the first ranks on a host resend the message
		//to the others on the same host (hopefully through local channels)	
		if(rank == MNB_GetResenderRank(env->localNode)){
			MNB_ResendMessageToNode(&message, env->localNode);
		}
		
		
		//quit messages mean, we can quit
		if(message.type == MNB_QUIT_MESSAGE){
			MNB_println("MNB_QUIT_MESSAGE message recieved");
			break;
			
		//for execute messages we have to create the comms and execute the test
		} else { 
			if(message.type == MNB_EXECUTE_MESSAGE){
				MNB_println("Execute message recieved");
				
				
				MNB_MessageBlock myMessageBlock;
				
				MPI_Comm testComm = MNB_CreateCommsFromMessage(&message, &myMessageBlock, testerNodesComm);
				
				//if we have found a test and a comm to work with
				if(testComm != MPI_COMM_NULL){
					MNB_ExecuteTestFromMessage(&myMessageBlock, testComm);
					MNB_println("MNB_ExecuteAsTester: test executed");
					MPI_Comm_free(&testComm);
					MNB_println("MNB_ExecuteAsTester: testComm freed");
				} else {
					MNB_println("MNB_ExecuteAsTester: no test to execute");
				}
			}
		}
	}
	//we can only get here by recievin an quit message
	//so now can free the testerNodesComm and guit
	if(*testerNodesComm != MPI_COMM_NULL){
		MPI_Comm_free(testerNodesComm);
	}
	free(testerNodesComm);
	
	
	//the scheduler will print out the tests, he could not schedule at the end
	//this barries makes shure the printout happens after all test results
	MPI_Barrier(MPI_COMM_WORLD);
}


/**
 * gets the information it needs from te messageBody (testName, sizes, ...)
 * and executes the test
 * 
 **/
static void MNB_ExecuteTestFromMessage(MNB_MessageBlock * messageBlock, MPI_Comm comm)
{
	testFunc function = MNB_GetTestByName(	messageBlock->body->testName, NULL);
											//~ messageBlock->body->testSource);
	int rank;
	MPI_Comm_rank(comm, &rank);
	
	if(function != NULL){
	
	//we have to execute the test, with all the repeats for each size
	for(int i = 0; i < messageBlock->header->numberOfSizes; i++){
		
		MNB_printf("%d:about to beginn execution of %s with size %d Byte, %d minRepeats, %d maxDuration, %d nodes, %d ranks\n",
					rank,
					messageBlock->body->testName,
					messageBlock->body->sizes[i], 
					messageBlock->body->minRepeats, 
					messageBlock->body->maxDuration, 
					messageBlock->header->numberOfNodes,
					messageBlock->header->numberOfRanks);
		
		//values to collect statistical information 
		double duration = 0, 
				minDuration = DBL_MAX, 
				maxDuration = 0, 
				avgDuration = 0;
				
		int repeats = 0;
		double startTime = MPI_Wtime();
		
		//while we haven't made enough repeats and aren't working longer then we are supposed to
		while(	(	repeats <= messageBlock->body->minRepeats) &&
				((MPI_Wtime() - startTime) <= messageBlock->body->maxDuration)){
			
			#ifdef DEBUG	
			if(rank == 0){
				printf("about to call the %s funktion\n", messageBlock->body->testName);
				}
				duration = function((messageBlock->body->sizes[i]) ,comm);
			if(rank == 0){
				printf("execution duration %f\n", duration);
			}
			#else
				duration = function((messageBlock->body->sizes[i]) ,comm);
			#endif
			
			//and now we are collecting statistic information
			if(duration < minDuration)
				minDuration = duration;
				
			if(duration > maxDuration)
				maxDuration = duration;
			
			avgDuration = ((avgDuration * repeats) + duration) / (repeats+ 1);
			
			repeats = repeats + 1;
			#ifdef DEBUG
				if(rank == 0){
					printf("executed test repeat:%d\n", repeats);
				}
			#endif
			}
			#ifdef DEBUG
				if(rank == 0){
					printf("executed test:%s\n", messageBlock->body->testName);
				}
			#endif
			
			double allMinDuration = 0, allMaxDuration = 0, allAvgDuration = 0;
			
			//now lets collect the statistacal information from all nodes
			MPI_Reduce(&minDuration, &allMinDuration, 1, MPI_DOUBLE, MPI_MIN, 0, comm);
			MPI_Reduce(&maxDuration, &allMaxDuration, 1, MPI_DOUBLE, MPI_MAX, 0, comm);
			MPI_Reduce(&avgDuration, &allAvgDuration, 1, MPI_DOUBLE, MPI_SUM, 0, comm);
			
			//now have to devide the avgDuration through the number of ransk in comm to get the real avg
			int numberOfRanks;
			MPI_Comm_size(comm, &numberOfRanks);
			allAvgDuration = allAvgDuration / numberOfRanks;
			
		
		//And now i can print the statistic information
		
		if(rank == 0){
			printf("Test:%s, Nodes:%d Ranks:%d, Size:%d byte, minDuration:%fs, maxDuration:%fs, avgDuration:%fs\n",
			messageBlock->body->testName, 
			messageBlock->header->numberOfNodes,
			messageBlock->header->numberOfRanks,
			messageBlock->body->sizes[i],
			allMinDuration,
			allMaxDuration,
			allAvgDuration);

		
		}
		
	}
	
	
	MNB_println("test executed");
	//MPI_Barrier(comm);
	} else { //if function == NULL
		printf("couldn't find %s function\n", messageBlock->body->testName);
	}

}







/**
 * creates a struct of appropriate size so a MNB_MessageBlockBody can actualy
 * be copied into by MPI_Recv
 * 
 **/
MNB_MessageBlockBody * MNB_AllocateMessageBlockBody(MNB_MessageBlockHeader * messageHeader){
	MNB_println("MNB_AllocateMessageBlockBody");
	MNB_MessageBlockBody * newMessageBody = malloc(sizeof(MNB_MessageBlockBody));
	newMessageBody->ranks = malloc(sizeof(int) * messageHeader->numberOfRanks);
	MNB_printf("pointer at the allocated memory: %p\n", newMessageBody->ranks);
	newMessageBody->sizes = malloc(sizeof(int) * messageHeader->numberOfSizes);
	//~ newMessageBody->testCollName = malloc(sizeof(char) * messageHeader->sizeOfTestCollName);
	newMessageBody->testName = malloc(sizeof(char) * messageHeader->sizeOfTestName);
	//~ newMessageBody->testSource = malloc(sizeof(char) * messageHeader->sizeOfTestSource );
	
	MNB_println("MNB_AllocateMessageBlockBody:done");
	
	return newMessageBody;
}


/**
 * Waits to recieve an MNB_MessageBlockBody and fills teh provided struct with this information
 * 
 * MNB_RecvMessage and MNB_RecvMessage could be rewritten to use a custom MPI_Datatype, but there is more to loose than to gain
 **/
void MNB_RecvMessage(MNB_Message * message, MNB_Environment * env)
{	
	MNB_println("MNB_RecvMessage");
	
	int sourceRank, rank;
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
	if(rank == MNB_GetResenderRank(env->localNode)){
		sourceRank = env->schedulerRank;
	} else {
		sourceRank = MNB_GetResenderRank(env->localNode);
	}
	
	//information about the size of the whole message
	MPI_Recv(&(message->type), 			1, MPI_INT, sourceRank, 1, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
	MPI_Recv(&(message->numberOfBlocks),1, MPI_INT, sourceRank, 1, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
	
	message->messageBlocks = malloc(sizeof(MNB_MessageBlock) * message->numberOfBlocks);
	
	
	for(int i = 0; i < message->numberOfBlocks; i++){
		MNB_println("MNB_RecvMessage:recieve Block");
		MNB_MessageBlockHeader * messageHeader = malloc(sizeof(MNB_MessageBlockHeader));
		
		//recieving the Header
		MPI_Recv(&(messageHeader->numberOfNodes),		1 , MPI_INT, sourceRank, 1, MPI_COMM_WORLD, MPI_STATUS_IGNORE );
		MPI_Recv(&(messageHeader->numberOfRanks),		1 , MPI_INT, sourceRank, 1, MPI_COMM_WORLD, MPI_STATUS_IGNORE );
		MPI_Recv(&(messageHeader->numberOfSizes),		1 , MPI_INT, sourceRank, 1, MPI_COMM_WORLD, MPI_STATUS_IGNORE );
		//~ MPI_Recv(&(messageHeader->sizeOfTestCollName),	1 , MPI_INT, sourceRank, 1, MPI_COMM_WORLD, MPI_STATUS_IGNORE );
		MPI_Recv(&(messageHeader->sizeOfTestName),		1 , MPI_INT, sourceRank, 1, MPI_COMM_WORLD, MPI_STATUS_IGNORE );
		//~ MPI_Recv(&(messageHeader->sizeOfTestSource),	1 , MPI_INT, sourceRank, 1, MPI_COMM_WORLD, MPI_STATUS_IGNORE );
						
		
		MNB_MessageBlockBody * messageBody = MNB_AllocateMessageBlockBody(messageHeader);
		
		
		// recieving the Body;
		MPI_Recv(messageBody->ranks, 		messageHeader->numberOfRanks, 	MPI_INT,  sourceRank, 1, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
		MPI_Recv(messageBody->sizes, 		messageHeader->numberOfSizes, 	MPI_INT,  sourceRank, 1, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
		//~ for(int i = 0; i< messageHeader->numberOfSizes; i++){
			//~ printf("Size recieved:%d\n", messageBody->sizes[i]);
		//~ }
		//~ MPI_Recv(messageBody->testCollName, messageHeader->sizeOfTestCollName,	MPI_CHAR, sourceRank, 1, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
		MPI_Recv(messageBody->testName, 	messageHeader->sizeOfTestName,		MPI_CHAR, sourceRank, 1, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
		//~ MPI_Recv(messageBody->testSource,	messageHeader->sizeOfTestSource,	MPI_CHAR, sourceRank, 1, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
		MPI_Recv(&(messageBody->minRepeats), 	1, 								MPI_INT,  sourceRank, 1, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
		MPI_Recv(&(messageBody->maxDuration), 	1, 								MPI_INT,  sourceRank, 1, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
		
		message->messageBlocks[i].body		= messageBody;
		message->messageBlocks[i].header	= messageHeader;	
	}
	MNB_println("MNB_RecvMessage:done");
}

/**
 * sends the messageBody to target rank
 * 
 * 
 **/
void MNB_SendMessage(MNB_Message * message, int targetRank)
{
	MNB_println("MNB_SendMessage");
	
	MPI_Send(&(message->type), 1, MPI_INT, targetRank, 1, MPI_COMM_WORLD);
	MPI_Send(&(message->numberOfBlocks), 1, MPI_INT, targetRank, 1, MPI_COMM_WORLD);
	
	for(int i = 0; i < message->numberOfBlocks; i++){
		MNB_println("MNB_SendMessage:send Block");
		MNB_MessageBlockHeader	* messageHeader = message->messageBlocks[i].header;
		MNB_MessageBlockBody 	* messageBody	= message->messageBlocks[i].body;
		
		//sending the Header
		MPI_Send(&(messageHeader->numberOfNodes),		1, MPI_INT, targetRank, 1, MPI_COMM_WORLD );
		MPI_Send(&(messageHeader->numberOfRanks),		1, MPI_INT, targetRank, 1, MPI_COMM_WORLD );
		MPI_Send(&(messageHeader->numberOfSizes),		1, MPI_INT, targetRank, 1, MPI_COMM_WORLD );
		//~ MPI_Send(&(messageHeader->sizeOfTestCollName),		1, MPI_INT, targetRank, 1, MPI_COMM_WORLD );
		MPI_Send(&(messageHeader->sizeOfTestName),		1, MPI_INT, targetRank, 1, MPI_COMM_WORLD );
		//~ MPI_Send(&(messageHeader->sizeOfTestSource),	1, MPI_INT, targetRank, 1, MPI_COMM_WORLD );
						
		
		
		
		// sending of the Body;
		MPI_Send(messageBody->ranks, 		messageHeader->numberOfRanks, 	MPI_INT,  targetRank, 1, MPI_COMM_WORLD);
		MPI_Send(messageBody->sizes, 		messageHeader->numberOfSizes, 	MPI_INT,  targetRank, 1, MPI_COMM_WORLD);
		//~ for(int i = 0; i< messageHeader->numberOfSizes; i++){
			//~ printf("Size send:%d\n", messageBody->sizes[i]);
		//~ }
		
		//~ MPI_Send(messageBody->testCollName, messageHeader->sizeOfTestCollName,	MPI_CHAR, targetRank, 1, MPI_COMM_WORLD);
		MPI_Send(messageBody->testName, 	messageHeader->sizeOfTestName,		MPI_CHAR, targetRank, 1, MPI_COMM_WORLD);
		//~ MPI_Send(messageBody->testSource,	messageHeader->sizeOfTestSource,	MPI_CHAR, targetRank, 1, MPI_COMM_WORLD);
		MPI_Send(&(messageBody->minRepeats), 	1, 								MPI_INT,  targetRank, 1, MPI_COMM_WORLD);
		MPI_Send(&(messageBody->maxDuration), 	1, 								MPI_INT,  targetRank, 1, MPI_COMM_WORLD);
	}
	
	

	
	MNB_println("MNB_SendMessage:done");
}



/**
 * !!! is supposed to be called by scheduler only !!!
 * 
 * Sends QUIT messages to all slave Ranks and ranks on scheduler node
 * execpt for the scheduler;
 * 
 **/
void MNB_SendQuitMessageToAll(MNB_Environment * env){
	MNB_println("MNB_SendQuitMessageToAll");
	
	//creating a message with the necessary fields sat
	MNB_Message message;
	message.type = MNB_QUIT_MESSAGE;
	message.numberOfBlocks = 0;
	
	//sending the message to all testerNode
	MNB_SendToResenders( &message, env); //NULL if for the unneccessary messageBody
	MNB_println("MNB_SendQuitMessageToAll:send to Testers");
	
	
	//and all the ranks on the schedulerNode
	if(!env->singleNodeExecution){ //sendToResenders would aready have done the job on a single node
		MNB_ResendMessageToNode(&message, env->schedulerNode );
	}
	MNB_println("MNB_SendQuitMessageToAll:allSend");
		
}



/**
 * Sends the Message to all other ranks on the same node
 * 
 * 
 **/
void MNB_ResendMessageToNode( MNB_Message * message, MNB_Node * node){
	MNB_println("MNB_ResendMessageToNode");
	assert(node 	!= NULL);
	assert(message	!= NULL);
	
	//we need our own rank, so we don't try to resend the message to ourself
	int myRank;
	MPI_Comm_rank(MPI_COMM_WORLD, &myRank);
	
	//simple loop to go through all of the ranks of the node
	struct MNB_ListElement * iterator = node->ranks->first; 
	while(iterator != NULL){
		int targetRank = *((int *)(iterator->data));
		//this funktion will be called with localNode as node parameter
		//so our own rank will probably also be in the list;
		if(myRank != targetRank){
			MNB_SendMessage(message, targetRank);
		}
		iterator = iterator->next;
	}
	MNB_println("MNB_ResendMessageToNode:send");
}


/**
 * creates a message containing all the information about the next round
 * of test and sends it to the resenders
 * 
 **/
void MNB_InformTesters(struct MNB_List * testList, struct MNB_List * nextRoundNodeLists, MNB_Environment * env, MPI_Comm * testerComm){
	MNB_println("MNB_InformTesters");
	
	//allocating the memory we need for the message
	MNB_Message * message = malloc(sizeof(MNB_Message));
	message->type = MNB_EXECUTE_MESSAGE;
	message->numberOfBlocks = testList->size;
	MNB_printf("%d messageBlocks will be created\n", message->numberOfBlocks);
	message->messageBlocks = malloc(sizeof(MNB_MessageBlock) * (message->numberOfBlocks));
	
	//and now we can fill the message blocks with information from testList and nextRoundNodeLists
	int i = 0;
	struct MNB_ListElement * testIterator = testList->first;
	struct MNB_ListElement * nodeListsIterator = nextRoundNodeLists->first;
	while(testIterator != NULL){	//it doesn't matter which one we take, the lists have to be the same length
		MNB_Test * test				= (MNB_Test *)testIterator->data;
		struct MNB_List * nodeList	= (struct MNB_List *)nodeListsIterator->data;
		
		//the actual filling of the information into the messageBlock
		MNB_FillMessageBlock(&(message->messageBlocks[i]), test, nodeList);
		
		//everyone has to move by one position
		testIterator		= testIterator->next;
		nodeListsIterator	= nodeListsIterator->next;
		i++;
		MNB_println("MessageBlock filled\n");
	}
	
	//now the message is complete and we can send it to the testers
	MNB_SendToResenders(message, env);
	
	
	//once the message is send we don't need it anymore
	MNB_FreeMessage(message);

		
	MNB_println("MNB_InformTesters:done");
}



/**
 * 
 * frees up the memory allocated by MNB_AllocateMessageBlockBody()
 **/
void MNB_FreeMessage(MNB_Message * message){
	MNB_println("MNB_FreeMessage");
	void * toDelete;
	for(int i = 0; i < message->numberOfBlocks; i++){
		MNB_printf("about to free messageBlock:%d\n", i);
		toDelete = message->messageBlocks[i].body->ranks;
		MNB_printf("to delete %p\n", toDelete);
		free(toDelete);
		toDelete = message->messageBlocks[i].body->sizes;
		free(toDelete);
		toDelete = message->messageBlocks[i].body->testName;
		free(toDelete);
		//~ free(message->messageBlocks[i].body->sizes);
		//~ free(message->messageBlocks[i].body->testName);
		//~ free(message->messageBlocks[i].body->testSource);
		free(message->messageBlocks[i].body);
		free(message->messageBlocks[i].header);
	}
	free(message->messageBlocks);
	free(message);
	MNB_println("MNB_FreeMessage:done");
}


void MNB_FillMessageBlock(MNB_MessageBlock * messageBlock, MNB_Test *test, struct MNB_List * nodeList ){
	MNB_println("MNB_FillMessageBlock");
	

	int sizesPosition = 0,  rankPosition = 0;
	
	messageBlock->header = malloc(sizeof(MNB_MessageBlockHeader));
	
	messageBlock->header->numberOfNodes		= MNB_GetNodeRequerment(test);
	messageBlock->header->numberOfRanks		= MNB_GetRankRequerment(test);
	messageBlock->header->numberOfSizes		= test->sizes->size;
	//~ messageBlock->header->sizeOfTestCollName	= strlen(test->name + 1)  ;
	messageBlock->header->sizeOfTestName		= strlen(test->test) + 1;
	messageBlock->header->sizeOfTestSource 	= 0;
	
	MNB_printf("sizeof testCollName: %d\n", messageBlock->header->sizeOfTestCollName);
	
	
	messageBlock->body = MNB_AllocateMessageBlockBody(messageBlock->header);
	strcpy(messageBlock->body->testName, test->test);
	//~ strcpy(messageBlock->body->testCollName, test->name);

	messageBlock->body->maxDuration = test->maxDuration;
	messageBlock->body->minRepeats  = test->minRepeats;

	
	
	//filling the sizes into messageBody
	struct MNB_ListElement * sizeIterator = test->sizes->first;
	MNB_println("MNB_FillMessageBlock:fill the sizes into the message");
	while(sizeIterator != NULL){
		messageBlock->body->sizes[sizesPosition] = MNB_SizeStringToInt(((MNB_Setting*)sizeIterator->data)->data);
		MNB_printf("Size %d filled into the message\n", messageBlock->body->sizes[sizesPosition]);
		sizeIterator = sizeIterator->next;
		sizesPosition = sizesPosition + 1;
	}
	
	
	MNB_println("MNB_FillMessageBlock:fill the ranks into the message");
	//and now its the ranks tern to be filled into messageBody
	//so we are goint through the nodes, 
	struct MNB_ListElement * nodeIterator = nodeList->first;
	while(nodeIterator != NULL){
		MNB_println("MNB_FillMessageBlock:beginn of the node loop");
		MNB_Node * node = (MNB_Node *)nodeIterator->data;
		struct MNB_ListElement * rankIterator = node->ranks->first;
		while(rankIterator != NULL){
			messageBlock->body->ranks[rankPosition] = *((int *)rankIterator->data);
			rankPosition++;
			rankIterator = rankIterator->next;
			//if this is a greedy scheduled test, we will have more ranks then we actualy need
			//so weh have to stop, when we have enough ranks
			if(rankPosition == messageBlock->header->numberOfRanks){
				break;
			}
		}
		
		
		nodeIterator = nodeIterator->next;
	}
	MNB_println("MNB_FillMessage:done");
	//~ return messageBlock;
}


/**
 * Sends the message to all ResenderRanks, so it can be redestributed localy
 * 
 **/
void MNB_SendToResenders( MNB_Message * message, MNB_Environment * env){
	MNB_println("MNB_SendToResenders");
	struct MNB_ListElement * iterator = env->testerNodes->first;
	while(iterator != NULL){
		MNB_Node * host = (MNB_Node *)iterator->data;
		int resenderRank = *((int *)host->ranks->first->data);
		MNB_printf("MNB_SendToResenders:about to send to:%d\n", resenderRank);
		
		MNB_SendMessage(message, resenderRank);
		iterator = iterator->next;
	}
	MNB_println("MNB_SendToResenders:done");
}



/**
 * creates all groups and comms described in the messageBlock
 * return the communicator containing the executing rank or
 * MPI_COMM_NULL
 * 
 **/
MPI_Comm MNB_CreateCommsFromMessage(MNB_Message * message, MNB_MessageBlock * messageBlock, MPI_Comm * parentComm){
	
	
	
	//~ MNB_MessageBlock myMessageBlock;
	MPI_Comm myComm = MPI_COMM_NULL;
	
	//create all groups and comms. remenber the one we need for our test
	for(int i = 0; i < message->numberOfBlocks ; i++){
		MNB_println("begin of the group/comm creation loop");
		
		MPI_Comm	* tmpComm = MNB_CreateCommFromMessageBlock(&(message->messageBlocks[i]), parentComm);
		
		
		
		#ifdef DEBUG
			int size;
			MPI_Comm_size(*parentComm, &size);
			printf("parentComm size:%d\n", size);
		#endif
		
		
		if(*tmpComm != MPI_COMM_NULL){
			
			*messageBlock = message->messageBlocks[i];
			myComm = *tmpComm;
		} else {
			MNB_println("I'm not part of the execution");
		}
		free(tmpComm);
		
	}
return myComm;
}
