#include"mnb-utils.h"
#include<ctype.h>


static void MNB_PrintSettingRec(MNB_Setting * typed_setting, int depth);


/**
 * tries to allocate size ammount of memory and informs the user on failure
 * if successfull also sets all bytes to 0
 * 
 **/
void * mallocWithCheck(long size){
  void * ret = malloc(size);
  if (ret == NULL){
    printf("Error could not malloc %lld of bytes\n", (long long int) size);
    MPI_Abort(1, MPI_COMM_WORLD);
  }
  memset(ret, 0, size);

  return ret;
}



/**
 * allocates the memory and sets default values for a test
 * 
 **/
MNB_Test * MNB_CreateTest(){
	MNB_println("CreateTest");
	MNB_Test * newTest = malloc(sizeof(MNB_Test));		
	newTest->sizes = MNB_CreateList();
	newTest->nodeConf = MNB_CreateList();
	newTest->minRepeats = 0;
	newTest->maxDuration = 0;
	return newTest;
	
}


/**
 * allocates the memory and setts the setting type to TYPE_NONE
 * 
 **/
MNB_Setting * MNB_CreateSetting(){
	MNB_println("CreateSetting");
	MNB_Setting * newSetting = malloc(sizeof(MNB_Setting));
	newSetting->type = TYPE_NONE;
	return newSetting;
}

/**
 * print the name, type and value of a setting
 * if setting is of type list, also prints all settings in the list
 * 
 **/
void MNB_PrintSettingRec(MNB_Setting * typedSetting, int depth)
{
	int depthIterator = depth;
	while(depthIterator != 0)
	{
		printf("%s", "\t");
		depthIterator = depthIterator - 1;
	}
	switch(typedSetting->type)
	{
		case TYPE_STRING:
			if(typedSetting->name != NULL){ 
				printf("string:%s:%s\n",typedSetting->name , (char *)typedSetting->data);
			} else {
				printf("string:%s\n",(char *)typedSetting->data);
			}
			break;
		case TYPE_INT:
			if(typedSetting->name != NULL){ 
				printf("int:%s:%d\n", typedSetting->name, *((int* )typedSetting->data));
			} else {
				printf("int:%d\n", *((int* )typedSetting->data));
			}
			break;
		case TYPE_NONE:
			printf("%s\n", "no type set");
			break;
		case TYPE_LIST:
			printf("Liste:%s\n", typedSetting->name);
			
			struct MNB_ListElement * iterator = ((struct MNB_List *)typedSetting->data)->first;
			
			while(iterator != NULL){
				MNB_PrintSettingRec((MNB_Setting *)iterator->data, depth + 1);
				iterator = iterator->next;
			}
			break;
		default:
			printf("%s\n", "unknown type");
		break;
		
	}
	
}


/**
 * creates a communicator for all the slaveRanks
 * this comm can be used to create the comms requered for the tests, 
 * without making the scheduling Rank part of the creation.
 * 
 **/
MPI_Comm * MNB_CreateTesterComm(MNB_Environment * env){
	MNB_println("MNB_CreateTesterComm");
	MPI_Comm * testerComm = malloc(sizeof(MPI_Comm));
	MPI_Group worldGroup, testerGroup;
	
	
	MPI_Comm_group(MPI_COMM_WORLD, &worldGroup);
	
	
	//if we are working on several nodes, we have to go though out the whole
	//schedulerNode from the worldGroup
	if(env->singleNodeExecution)
	{
		MNB_println("single Node execution:create tester comm");
		//with single Host execution we just have to though out the scheduler
		MPI_Group_excl(worldGroup, 1, &(env->schedulerRank), &testerGroup);
		testerGroup = worldGroup;
		
	} else {
		MNB_println("multi Node execution:create tester comm");
		//since we are storing our ranks as a MNB_List, we have to convert
		//the list to an array so we can pass it on to MPI_Group_excl
		int * ranksOnSchedulerNode = malloc(sizeof(int) * (env->schedulerNode->ranks->size + 1));
		MNB_printf("multi Host execution: %d ranks to exclude \n", env->schedulerNode->ranks->size);
		struct MNB_ListElement * iterator = env->schedulerNode->ranks->first;
		
		//we have removed schedulerRank from the schedulerNode
		//so we have to put it into the new array by hand
		ranksOnSchedulerNode[0] = env->schedulerRank;
		int ranksPosition = 1;
		while(iterator != NULL){
			ranksOnSchedulerNode[ranksPosition] = *((int*)iterator->data);
			MNB_printf("multi Node execution: %d excluded from testerranks\n",ranksOnSchedulerNode[ranksPosition] );
			ranksPosition = ranksPosition + 1;
			iterator = iterator->next;
		}
		
		//now we have the rank list, so we can exlude them all from the new groub
		MPI_Group_excl(worldGroup, env->schedulerNode->ranks->size + 1, //again, the +1 is for the scheduler rank
						ranksOnSchedulerNode, &testerGroup);
		
		//we dont need the ranks array anymore 
		free(ranksOnSchedulerNode);
		
	} 
	
	
	MPI_Comm_create(MPI_COMM_WORLD, testerGroup, testerComm);
	MNB_printf("%s\n", "create tester comm:done");
	
	//some debug information
	#ifdef DEBUG
		MPI_Barrier(MPI_COMM_WORLD);
		printf("%s\n", "create tester comm:MPI_COMM_WORLD barrier passed");
		
		if(*testerComm != MPI_COMM_NULL){
			int globalRank, testerRank;
			MPI_Comm_rank(MPI_COMM_WORLD, &globalRank);
			MPI_Comm_rank(*testerComm, &testerRank);
			
			printf("%d:create tester comm:TESTER barrier passed. tester Rank %d\n", globalRank, testerRank);
		}
	#endif
	
	MNB_println("MNB_CreateTesterComm:done");
	return testerComm;
	
}

/**
 * creates a MPI_Group using information from messageBlock
 * 
 * has to be called by all members of MPI_Comm
 * 
 **/
MPI_Comm * MNB_CreateCommFromMessageBlock(MNB_MessageBlock * messageBlock, MPI_Comm * comm){
	MNB_println("MNB_CreateCommFromMessageBlock");
	
	MNB_println("MNB_CreateCommFromMessageBlock:global Ranks in group:");	
	for(int j = 0; j < messageBlock->header->numberOfRanks; j++){
						MNB_printf("%d ", messageBlock->body->ranks[j]);
					}
	MNB_println(" ");
	
	// the actual creation starts here
	MPI_Group * localGroup = malloc(sizeof(MPI_Group));
	MPI_Comm	* localComm = malloc(sizeof(MPI_Comm));
	int * localRanks = malloc(sizeof(int) * messageBlock->header->numberOfRanks);
	
	MNB_println("MNB_CreateCommFromMessageBlock:memory allocated");
	
	MPI_Group worldGroup, slaveGroup;
	MPI_Comm_group(MPI_COMM_WORLD, &worldGroup);
	MPI_Comm_group(*comm, &slaveGroup);
	
	MNB_println("MNB_CreateCommFromMessageBlock:groups created");
	
	//we have to convert the ranks of MPI_COMM_WORLD, which the schedular is using to ranks of our local comm
	MPI_Group_translate_ranks(worldGroup, 
								messageBlock->header->numberOfRanks, 
								messageBlock->body->ranks, 
								slaveGroup, 
								localRanks);
	MNB_printf("MNB_CreateCommFromMessageBlock:ranks translated");
							
	//creation of the group of which we need the comm
	MPI_Group_incl(	slaveGroup, 
					messageBlock->header->numberOfRanks, 
					localRanks, 
					localGroup);
					
	
	MNB_println("MNB_CreateCommFromMessageBlock:testerRanks in group:");				
	for(int j = 0; j < messageBlock->header->numberOfRanks; j++){
						MNB_printf("%d ", localRanks[j]);
					}
	MNB_println(" ");
					
	
	
	MPI_Comm_create(*comm, *localGroup, localComm);
	

	MNB_println("MNB_CreateCommFromMessageBlock:Comm created from Group");
	
	
	#ifdef DEBUG
		if(*localComm != MPI_COMM_NULL){
			int globalRank, localRank, localSize;
			MPI_Comm_rank(MPI_COMM_WORLD, &globalRank);
			MPI_Comm_rank(*localComm, &localRank);
			MPI_Comm_size(*localComm, &localSize);
			printf("MNB_CreateCommFromMessageBlock:global Rank:%d, localRank %d, localSize %d\n", globalRank, localRank, localSize);
		}
	#endif
	
	
	//we don't need the raks as an array anymore
	free(localRanks);	
	//we have the comm, we don't need the group anymore
	MPI_Group_free(localGroup);
	free(localGroup);	
	
	
	return localComm;
}


/**
 * this is funktion just calls MNB_PrintSettingRec as 0 as the depth parameter
 * its's only use is to hide the second parameter
 * 
 * apparantly c doesn't support dalualt values for parameter
 **/
void MNB_PrintSetting(MNB_Setting * typedSetting){
	MNB_println("MNB_PrintSetting");
	MNB_PrintSettingRec(typedSetting, 0);
}



/**
 * prints the name,  size an node information of a test
 * 
 * 
 **/
void MNB_PrintTest(MNB_Test * test){
	MNB_println("MNB_PrintTest");
	
	//~ printf("name: %s\n", test->name);
	printf("test: %s\n", test->test);
	
	
	printf("%s\n", "sizes:");
	
	struct MNB_ListElement * iterator = test->sizes->first;
	while(iterator != NULL){
		MNB_PrintSetting((MNB_Setting *)iterator->data);
		iterator = iterator->next;
	}
	
	printf("node conf: %d nodes, %d ranks\n", 	MNB_GetNodeRequerment(test),
												MNB_GetRankRequerment(test));
	
	printf("minRepeats:  %d\n", test->minRepeats);
	printf("maxDuration: %d\n", test->maxDuration);
	
	
}


/**
 * 
 * Creates an schallow copy of an MNB_Test
 * 
 **/
MNB_Test * MNB_CopyTest(MNB_Test * run_orig)
{
	MNB_Test * new_run = MNB_CreateTest();
	new_run->name = run_orig->name;
	new_run->nodeConf = run_orig->nodeConf;
	new_run->sizes = run_orig->sizes;
	new_run->test = run_orig->test;
	new_run->minRepeats = run_orig->minRepeats;
	new_run->maxDuration = run_orig->maxDuration;
	return new_run;
	
}


/**
 * converts sizes provided by the configurationFile to integervalues of 
 * bytes
 * 
 **/
int MNB_SizeStringToInt(char * size){
	int value = 1, scaleMultiplier = 1;
	
	//10 might be needed, in case we will allow other scale values lates
	char scaleString[10];
	sscanf(size, "%d%s", &value, scaleString);
		
	char scaleChar = tolower(scaleString[0]);
	
	if(scaleChar == 'k'){
		scaleMultiplier = 1024;
	} else { if(scaleChar == 'm'){
		scaleMultiplier = 1024*1024;
	} else { if(scaleChar == 'g'){
		scaleMultiplier = 1024*1024*1024;
	} 
	   else {
		   printf("couldn't convert %s to a usable datasize, using %d byte for tests\n", size, value * scaleMultiplier); 
	   } }
	}
	MNB_printf("converted %s to datasize, using %d byte for tests\n", size, value * scaleMultiplier);
	return value * scaleMultiplier;
}



/**
 * rank 0 prints an welcome message for the user with some information about the environment
 * 
 **/
void MNB_PrintWelcome(MNB_Environment * env){
	int rank;
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
	if(rank == 0){
		printf("\n%s",	"************************************************************\n");
				printf(	"*             Working with %2d ranks on %2d nodes            *\n", env->numberOfRanks, env->numberOfNodes);
				
				printf(	"*             %d is schedueler : %s                *\n", env->schedulerRank, ((env->singleNodeExecution)?("single Node"):("multi Node ")));
		
		printf("%s",	"************************************************************\n");
	}
}
